/*
 *	dnsutl - utilities to make DNS easier to configure
 *	Copyright (C) 2001, 2006 Peter Miller;
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 * MANIFEST: functions to filter dhcp entries
 */

#include <ac/ctype.h>
#include <ac/string.h>
#include <ac/stdlib.h>

#include <arglex.h>
#include <error.h>
#include <filter.h>
#include <mem.h>
#include <output.h>
#include <srrf/dhcp.h>
#include <srrf/reader.h>
#include <srrf/origin.h>
#include <symtab.h>


int verbose;


static void
load_the_defaults(symtab_ty *stp, string_ty *key, void *data, void *arg)
{
	symtab_ty	*stp2;
	void		*data2;

	(void)stp;
	stp2 = arg;
	data2 = symtab_query(stp2, key);
	if (!data2)
		symtab_assign(stp2, key, data);
}


static srrf_t *
dhcp_hardware_ethernet_from_ether_a(srrf_t *rp)
{
	srrf_t		*rp2;

	rp2 = srrf_alloc();
	rp2->name = str_copy(rp->name);
	rp2->class = srrf_class_by_name("dhcp");
	rp2->type = srrf_type_by_name(rp2->class, "hardware-ethernet");
	strlist_append(&rp2->arg, rp->name);
	rp2->file_name = str_copy(rp->file_name);
	rp2->line_number = rp->line_number;
	return rp2;
}


static srrf_t *
dhcp_fixed_address_from_in_a(srrf_t *rp)
{
	srrf_t		*rp2;

	rp2 = srrf_alloc();
	rp2->name = str_copy(rp->name);
	rp2->class = srrf_class_by_name("dhcp");
	rp2->type = srrf_type_by_name(rp2->class, "fixed-address");
	strlist_append(&rp2->arg, rp->name);
	rp2->file_name = str_copy(rp->file_name);
	rp2->line_number = rp->line_number;
	return rp2;
}


static int
cmp(const void *va, const void *vb)
{
	string_ty	*a;
	string_ty	*b;

	a = *(string_ty **)va;
	b = *(string_ty **)vb;
	return strcmp(a->str_text, b->str_text);
}


void
filter(const char *infile, const char *outfile)
{
	srrf_list_ty	*rlp;
	size_t		j, k;
	srrf_class_ty	*dhcp_class;
	srrf_class_ty	*ether_class;
	srrf_class_ty	*in_class;
	srrf_type_ty	*ether_a_type;
	srrf_type_ty	*in_a_type;
	srrf_type_ty	*in_cname_type;
	symtab_ty	*defaults_stp;
	int		number_of_errors;
	dhcp_aux_ty	aux;
	srrf_t		*rp;
	srrf_t		*rp2;
	string_ty	*name;

	/*
	 * read the input file
	 */
	rlp = srrf_reader(infile, verbose, 1);

	/*
	 * open the output file
	 */
	output_open(outfile);

	/*
	 * find the necessary type pointer
	 */
	dhcp_class = srrf_class_by_name("dhcp");
	assert(dhcp_class);
	ether_class = srrf_class_by_name("ether");
	assert(ether_class);
	in_class = srrf_class_by_name("in");
	assert(in_class);
	ether_a_type = srrf_type_by_name(ether_class, "a");
	assert(ether_a_type);
	in_a_type = srrf_type_by_name(in_class, "a");
	assert(in_a_type);
	in_cname_type = srrf_type_by_name(in_class, "cname");
	assert(in_cname_type);

	/*
	 * warning comment at the top of the file
	 */
	output_printf("# Do not edit this file.  It is generated\n");
	output_printf
	(
		"# using a '%s %s %s' command.\n",
		progname,
		(infile ? infile : "-"),
		(outfile ? outfile : "-")
	);

	/*
	 * scan the resources looking for ether addresses
	 */
	aux.in_a_stp = symtab_alloc(5);
	aux.ether_a_stp = symtab_alloc(5);
	for (j = 0; j < rlp->nrecords; ++j)
	{
		rp = rlp->record[j];
		if (rp->type == ether_a_type)
			symtab_assign(aux.ether_a_stp, rp->name, rp);
		if (rp->type == in_a_type)
			symtab_assign(aux.in_a_stp, rp->name, rp);
		if (rp->type == in_cname_type)
			symtab_assign(aux.in_a_stp, rp->name, rp);
	}

	/*
	 * scan the resources looking for hosts
	 */
	defaults_stp = symtab_alloc(5);
	number_of_errors = 0;
	for (j = 0; j < rlp->nrecords; ++j)
	{
		symtab_ty	*stp;
		int		number_of_params;

		rp = rlp->record[j];
		if
		(
			rp->class == dhcp_class
		&&
			rp->name->str_text[0] == '*'
		&&
			rp->name->str_text[1] == '.'
		)
		{
			/*
			 * set a defailt
			 */
			name = str_from_c(rp->type->name);
			symtab_assign(defaults_stp, name, rp);
			str_free(name);
		}
		if (rp->type != in_a_type)
			continue;

		/*
		 * collect all of the associated dhcp records
		 */
		number_of_params = 0;
		stp = symtab_alloc(5);
		for (k = 0; k < rlp->nrecords; ++k)
		{
			rp2 = rlp->record[k];
			if (!str_equal(rp2->name, rp->name))
				continue;
			if (rp2->class != dhcp_class)
				continue;

			name = str_from_c(rp2->type->name);
			symtab_assign(stp, name, rp2);
			str_free(name);
			++number_of_params;
		}

		/*
		 * print the records
		 */
		if
		(
			number_of_params > 0
		||
			symtab_query(aux.ether_a_stp, rp->name)
		)
		{
			strlist_ty	key;

			symtab_walk(defaults_stp, load_the_defaults, stp);
			name = srrf_absolute_to_relative(rp->name);
			output_printf("host %s {\n", name->str_text);
			str_free(name);

			/*
			 * generate an "fixed-address" entry
			 */
			rp2 = dhcp_fixed_address_from_in_a(rp);
			name = str_from_c(rp2->type->name);
			symtab_assign(stp, name, rp2);
			str_free(name);
			++number_of_params;

			/*
			 * generate a "hardware ethernet" entry
			 */
			rp2 = dhcp_hardware_ethernet_from_ether_a(rp);
			name = str_from_c(rp2->type->name);
			symtab_assign(stp, name, rp2);
			str_free(name);
			++number_of_params;

			/*
			 * Order they keys alphabetically
			 */
			symtab_keys(stp, &key);
			qsort
			(
				key.string,
				key.nstrings,
				sizeof(key.string[0]),
				cmp
			);

			/*
			 * print the keys out in order
			 */
			for (k = 0; k < key.nstrings; ++k)
			{
				rp2 = symtab_query(stp, key.string[k]);
				assert(rp2);
				srrf_invoke_aux1(rp2, &aux);
			}
			strlist_free(&key);
			symtab_free(stp);
			output_printf("}\n");
		}
	}
	symtab_free(defaults_stp);
	srrf_list_free(rlp);

	/*
	 * close the output file
	 */
	output_close();
}
